example_headers = ["Inorg_arb_examplepy_plugin_0_fun_count_drp_0.02",
"_N?_examplepy_plugin_0_False_molarity_drp_0.02",
"Ox_N?_examplepy_plugin_0_True_molarity_drp_0.02",
"Org_arb_examplepy_plugin_0_fun_count_drp_0.02",
"_N?_examplepy_plugin_0_True_molarity_drp_0.02",
"_arb_examplepy_plugin_0_dull_molarity_drp_0.02",
"Ox_arb_examplepy_plugin_0_fun_molarity_drp_0.02",
"Org_N?_examplepy_plugin_0_False_molarity_drp_0.02",
"Solv_N?_examplepy_plugin_0_False_molarity_drp_0.02",
"Org_fs_examplepy_plugin_0_1_count_drp_0.02",
"Solv_fs_examplepy_plugin_0_1_count_drp_0.02",
"Solv_fs_examplepy_plugin_0_3_count_drp_0.02",
"Inorg_N?_examplepy_plugin_0_True_molarity_drp_0.02",
"Org_arb_examplepy_plugin_0_dull_count_drp_0.02",
"Org_N?_examplepy_plugin_0_True_molarity_drp_0.02",
"Org_arb_examplepy_plugin_0_dull_molarity_drp_0.02",
"Inorg_N?_examplepy_plugin_0_False_count_drp_0.02",
"_fs_examplepy_plugin_0_3_count_drp_0.02",
"Solv_arb_examplepy_plugin_0_fun_molarity_drp_0.02",
"Solv_fs_examplepy_plugin_0_2_molarity_drp_0.02",
"Solv_arb_examplepy_plugin_0_fun_count_drp_0.02",
"Ox_fs_examplepy_plugin_0_3_count_drp_0.02",
"Inorg_fs_examplepy_plugin_0_1_count_drp_0.02",
"Ox_arb_examplepy_plugin_0_fun_count_drp_0.02",
"_fs_examplepy_plugin_0_2_count_drp_0.02",
"Ox_fs_examplepy_plugin_0_1_molarity_drp_0.02",
"_N?_examplepy_plugin_0_True_count_drp_0.02",
"Ox_N?_examplepy_plugin_0_False_count_drp_0.02",
"Inorg_fs_examplepy_plugin_0_2_molarity_drp_0.02",
"Inorg_fs_examplepy_plugin_0_2_count_drp_0.02",
"Inorg_N?_examplepy_plugin_0_True_count_drp_0.02",
"_arb_examplepy_plugin_0_fun_molarity_drp_0.02",
"Solv_arb_examplepy_plugin_0_dull_molarity_drp_0.02",
"Inorg_arb_examplepy_plugin_0_dull_molarity_drp_0.02",
"Inorg_arb_examplepy_plugin_0_dull_count_drp_0.02",
"Ox_arb_examplepy_plugin_0_dull_count_drp_0.02",
"_arb_examplepy_plugin_0_fun_count_drp_0.02",
"Ox_fs_examplepy_plugin_0_1_count_drp_0.02",
"Ox_fs_examplepy_plugin_0_3_molarity_drp_0.02",
"Inorg_N?_examplepy_plugin_0_False_molarity_drp_0.02",
"Org_fs_examplepy_plugin_0_3_count_drp_0.02",
"Solv_arb_examplepy_plugin_0_dull_count_drp_0.02",
"Ox_fs_examplepy_plugin_0_2_molarity_drp_0.02",
"_fs_examplepy_plugin_0_1_count_drp_0.02",
"Org_arb_examplepy_plugin_0_fun_molarity_drp_0.02",
"Ox_fs_examplepy_plugin_0_2_count_drp_0.02",
"Org_fs_examplepy_plugin_0_1_molarity_drp_0.02",
"Solv_fs_examplepy_plugin_0_2_count_drp_0.02",
"_fs_examplepy_plugin_0_3_molarity_drp_0.02",
"_fs_examplepy_plugin_0_1_molarity_drp_0.02",
"Ox_arb_examplepy_plugin_0_dull_molarity_drp_0.02",
"Org_N?_examplepy_plugin_0_True_count_drp_0.02",
"Solv_N?_examplepy_plugin_0_False_count_drp_0.02",
"Org_fs_examplepy_plugin_0_3_molarity_drp_0.02",
"_fs_examplepy_plugin_0_2_molarity_drp_0.02",
"Inorg_arb_examplepy_plugin_0_fun_molarity_drp_0.02",
"Inorg_fs_examplepy_plugin_0_3_molarity_drp_0.02",
"Solv_fs_examplepy_plugin_0_1_molarity_drp_0.02",
"Ox_N?_examplepy_plugin_0_False_molarity_drp_0.02",
"Solv_N?_examplepy_plugin_0_True_molarity_drp_0.02",
"Solv_N?_examplepy_plugin_0_True_count_drp_0.02",
"Ox_N?_examplepy_plugin_0_True_count_drp_0.02",
"_N?_examplepy_plugin_0_False_count_drp_0.02",
"Org_fs_examplepy_plugin_0_2_molarity_drp_0.02",
"Org_fs_examplepy_plugin_0_2_count_drp_0.02",
"_arb_examplepy_plugin_0_dull_count_drp_0.02",
"Inorg_fs_examplepy_plugin_0_3_count_drp_0.02",
"Org_N?_examplepy_plugin_0_False_count_drp_0.02",
"Inorg_fs_examplepy_plugin_0_1_molarity_drp_0.02",
"Solv_fs_examplepy_plugin_0_3_molarity_drp_0.02",
]





from AbstractRepairer import AbstractRepairer
from CategoricRepairer import Repairer as CategoricRepairer
from NumericRepairer import Repairer as NumericRepairer

class Repairer(AbstractRepairer):
  def __init__(self, *args, **kwargs):
    super(Repairer, self).__init__(*args, **kwargs)
    
    if self.kdd:
        self.repairer = CategoricRepairer(*args,**kwargs) 
    else:
        if all(isinstance(row[self.feature_to_repair],float) for row in self.all_data):
          self.repairer = NumericRepairer(*args, **kwargs)
        elif all(isinstance(row[self.feature_to_repair],int) for row in self.all_data):
          self.repairer = NumericRepairer(*args, **kwargs)
        elif all(isinstance(row[self.feature_to_repair],str) for row in self.all_data):
          self.repairer = CategoricRepairer(*args, **kwargs)

  def repair(self, data_to_repair):
    return self.repairer.repair(data_to_repair)



import csv

def load_data():
  filename = "test_data/RicciDataMod.csv"
  with open(filename) as f:
    reader = csv.reader(f)
    data = [row for row in reader]
    headers = data.pop(0)

    correct_types = [str, float, int, str, float, str]

    for i, row in enumerate(data):
      for j, correct_type in enumerate(correct_types):
        data[i][j] = correct_type(row[j])

    train = data[:len(data)//2]
    test = data[len(data)//2:]

  return headers, train, test


def test():
  headers, train, test = load_data()
  print("load_data unpacks correctly? -- ", (headers != None and train != None and test != None))

  correct_types = [str, float, int, str, float, str]
  gathered_types = []
  for i, header in enumerate(headers):
    if all( isinstance(row[i],float) for row in train + test ):
      gathered_types.append(float)
    elif all( isinstance(row[i],int) for row in train + test ):
      gathered_types.append(int)
    elif all( isinstance(row[i],str) for row in train + test ):
      gathered_types.append(str)

  print("load_data types are correct? -- ", gathered_types == correct_types)


if __name__=="__main__":
  test()





import os

datasets_dir = os.path.dirname(__file__)

preloaded = {
   "adult":    {"filepath": datasets_dir+"/adult.csv",
                "testdata": datasets_dir+"/adult.test.csv",
                "correct_types": [int, str, int, str, int, str, str, str,
                                  str,str, int, int, int, str, str],
                "train_percentage": None,
                "response_header": "income-per-year",
                "features_to_ignore": []},

   "diabetes": {"filepath": datasets_dir+"/pima-indians-diabetes.csv",
                "testdata": None,
                "correct_types": [int, float, float, float, float,
                                  float, float, int, str],
                "train_percentage": 1.0/2.0,
                "response_header": "Class",
                "features_to_ignore": []},

   "ricci":    {"filepath": datasets_dir+"/RicciDataMod.csv",
                "testdata": None,
                "correct_types": [str,float,int,str,float,str],
                "train_percentage": 1.0/2.0,
                "response_header": "Class",
                "features_to_ignore":[]},

   "german":   {"filepath": datasets_dir+"/german_categorical.csv",
                "testdata": None,
                "correct_types": [str, int, str, str, int, str, str, int, str, str, int,
                                  str, int, str, str, str, int, str, int, str, str, str],
                "train_percentage": 2.0/3.0,
                "response_header": "class",
                "features_to_ignore": ["age"]},

   "glass":    {"filepath": datasets_dir+"/glass.csv",
                "testdata": None,
                "correct_types": [int, float, float, float, float, float, float, float,
                                  float, float, str],
                "train_percentage": 1.0/2.0,
                "response_header": "type of glass",
                "features_to_ignore": []},

   "sample":   {"filepath": datasets_dir+"/sample.csv",
                "testdata": None,
                "correct_types": [int, int, int, int, float, str],
                "train_percentage":2.0/3.0,
                "response_header": "Outcome",
                "features_to_ignore": []},

   "DRP":      {"filepath": "test_data/DRP_nature_train.arff",
                "testdata": "test_data/DRP_nature_test.arff",
                "correct_types": [float]*2+[str,float,str]+[float]*162+[str]*58+[float]*48+[str],
                "train_percentage": 2.0/3.0,
                "response_header": "outcome",
                "features_to_ignore": []}

  }



def accuracy(conf_matrix):
  """
  Given a confusion matrix, returns the accuracy.
  Accuracy Definition: http://research.ics.aalto.fi/events/eyechallenge2005/evaluation.shtml
  """
  total, correct = 0.0, 0.0
  for true_response, guess_dict in conf_matrix.items():
    for guess, count in guess_dict.items():
      if true_response == guess:
        correct += count
      total += count
  return correct/total

def BCR(conf_matrix):
  """
  Given a confusion matrix, returns Balanced Classification Rate.
  BCR is (1 - Balanced Error Rate).
  BER Definition: http://research.ics.aalto.fi/events/eyechallenge2005/evaluation.shtml
  """
  parts = []
  for true_response, guess_dict in conf_matrix.items():
    error = 0.0
    total = 0.0
    for guess, count in guess_dict.items():
      if true_response != guess:
        error += count
      total += count
    parts.append(error/total)
  BER = sum(parts)/len(parts)
  return 1 - BER

def get_conf_matrix(prediction_tuples):
  # Produce a confusion matrix in a dictionary format from those predictions.
  conf_table = {}
  for actual, guess in prediction_tuples:
    guess = convert_to_type(actual, guess) # ... Since file-reading changes types.

    if not actual in conf_table:
      conf_table[actual] = {}

    if not guess in conf_table[actual]:
      conf_table[actual][guess] = 1
    else:
      conf_table[actual][guess] += 1

  return conf_table

def convert_to_type(actual, guess):
  if type(actual) == bool:
    guess = guess=="True"
  else:
    actual_type = type(actual)
    guess = actual_type(guess)
  return guess


def test():
  test_conf_matrix()
  test_measurers()

def test_conf_matrix():
  pred_tuples = [(1,1),(1,1),(2,2),(3,3),(1,3),(3,1)]
  correct_conf_matrix = {1:{1:2, 3:1}, 2:{2:1}, 3:{1:1, 3:1}}
  conf_matrix = get_conf_matrix(pred_tuples)
  print "confusion matrix correct? -- ", conf_matrix == correct_conf_matrix

def test_measurers():
  conf_matrix = {"A":{"A":10}, "B":{"A":5,"B":5}}
  print "measurements -- accuracy correct? -- ", accuracy(conf_matrix) == 0.75
  print "measurements -- 1-BER correct? -- ", BCR(conf_matrix) == 0.75

  conf_matrix = {"A":{"A":10}}
  print "measurements -- accuracy correct? -- ", accuracy(conf_matrix) == 1.0
  print "measurements -- 1-BER correct? -- ", BCR(conf_matrix) == 1.0

  conf_matrix = {"A":{"A":95}, "B":{"A":5}}
  print "measurements -- accuracy correct? -- ", accuracy(conf_matrix) == 0.95
  print "measurements -- 1-BER correct? -- ", BCR(conf_matrix) == 0.50

if __name__=="__main__":
  test()



import Orange
import os,csv

def CN2_learner(trainfile, testfile, output_dir, beam_width, min_covered_examples, max_rule_length, influence_scores):
  print("Setting up CN2 Learner")
  # format data for classification
  training_data = Orange.data.Table.from_file(trainfile)
  # set the learner
  learner = Orange.classification.rules.CN2Learner()
  # set the rule evaluator to be Laplace
  LaplaceEvaluator = Orange.classification.rules.LaplaceAccuracyEvaluator()
  learner.rule_finder.quality_evaluator = LaplaceEvaluator
  # set the number of solution steams considered at one time
  learner.rule_finder.search_algorithm.beam_width = beam_width
  # continuous value space is constrained to reduce computation time
  learner.rule_finder.search_strategy.constrain_continuous = True
  # set the minimum number of examples a found rule must cover to be considered
  learner.rule_finder.general_validator.min_covered_examples = min_covered_examples
  # set the maximum number of selectors (conditions) found rules may combine
  learner.rule_finder.general_validator.max_rule_length = max_rule_length

  print(("Learning rule list for {}".format(trainfile)))
  # learn a rule list for the data
  classifier = learner(training_data)

  print("Writing rules to file")
  # write rules to file
  rulesfile = "{}/rules.csv".format(output_dir)
  with open(rulesfile, 'w') as csvfile:
    rules = csv.writer(csvfile)
    # Create rules file from repaired data
    rules.writerow(["Label","Rules","Quality","Score"])
    for rule_num, rule in enumerate(classifier.rule_list):
      # calculate influence score
      domain = rule.domain.attributes
      selectors = rule.selectors
      score = sum([float(influence_scores[domain[s.column].name]) for s in selectors])
      # write rule details to file
      rules.writerow([rule_num, str(rule).strip(' '), rule.quality, score])

  print("Evaluating Model")
  test_data = Orange.data.Table.from_file(testfile)
  res = Orange.evaluation.TestOnTestData(training_data, test_data, [learner])
  accuracy = Orange.evaluation.scoring.CA(res)
  AUC = Orange.evaluation.scoring.AUC(res)
  return rulesfile, accuracy, AUC



from copy import deepcopy

def get_median(values, kdd):
  """
  Given an unsorted list of numeric values, return median value (as a float).
  Note that in the case of even-length lists of values, we apply the value to
  the left of the center to be the median (such that the median can only be
  a value from the list of values).
  Eg: get_median([1,2,3,4]) == 2, not 2.5.
  """

  if not values:
    raise Exception("Cannot calculate median of list with no values!")

  sorted_values = deepcopy(values)
  sorted_values.sort() # Not calling `sorted` b/c `sorted_values` may not be list.

  if kdd:
    return sorted_values[len(values)/2]
  else:
    if len(values) % 2 == 0:
      return sorted_values[len(values)/2-1]
    else:
      return sorted_values[len(values)/2]


def test():
  test_median()

def test_median():
  feature_values = [4,1,3,2]
  correct_median = 2
  # not KDD repair
  print "Regular median value is correct?", get_median(feature_values, False) == correct_median
  # KDD repair
  correct_median = 3
  print "KDD median value is correct?", get_median(feature_values, True) == correct_median

if __name__=="__main__":
  test()



#!/usr/bin/env python

from setuptools import setup, find_packages
from setuptools.command.develop import develop
from setuptools.command.install import install
from codecs import open
import os, sys 

here = os.path.abspath(os.path.dirname(__file__))

if sys.version_info[0] == 3:
    source_dir = '.'
else:
    source_dir = 'python2_source'

with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
  long_description = f.read()

NAME ='BlackBoxAuditing'
VERSION = '0.1.21'

DESCRIPTION = 'Sample Implementation of Gradient Feature Auditing (GFA)'
LONG_DESCRIPTION = long_description

URL = 'https://github.com/algofairness/BlackBoxAuditing'
AUTHOR ='Philip Adler, Casey Falk, Sorelle A. Friedler, Gabriel Rybeck, Carlos Scheidegger, Brandon Smith, Suresh Venkatasubramanian, Michael Feldman, John Moeller, Derek Roth, Charlie Marx'
AUTHOR_EMAIL='fairness@haverford.edu'
LICENSE='Apache 2.0'

CLASSIFIERS = [
  'Development Status :: 4 - Beta',
  'Intended Audience :: Developers',
  'Topic :: Software Development',
  'License :: OSI Approved :: Apache Software License',
  'Programming Language :: Python :: 2.7',
  'Programming Language :: Python :: 3.0',
]
KEYWORDS = 'algorithmic fairness'

PACKAGES = find_packages()
PACKAGE_DATA = {
  'BlackBoxAuditing': ['*.sh','r_plotting/*.Rmd', 'repair_tests/*.csv', 'MATLAB_code/*', 'weka.jar'],
  'BlackBoxAuditing.test_data': ['*.csv', '*.arff']
}
INCLUDE_PACKAGE_DATA = True
PACKAGE_DIR = {
  '': source_dir,
}

INSTALL_REQUIRES = [
  'tensorflow',
  'networkx',
  'matplotlib',
  'Orange3']

CMDCLASS = {
}

ENTRY_POINTS = {
  'console_scripts': [
    'BlackBoxAuditing-test = BlackBoxAuditing.test:test',
    'BlackBoxAuditing-repair = BlackBoxAuditing.repair:main',
  ],
}

setup(
  name=NAME,
  version=VERSION,
  description=DESCRIPTION,
  long_description=LONG_DESCRIPTION,
  url=URL,
  author=AUTHOR,
  author_email=AUTHOR_EMAIL,
  license=LICENSE,
  classifiers=CLASSIFIERS,
  keywords=KEYWORDS,
  packages=PACKAGES,
  package_data=PACKAGE_DATA,
  include_package_data=INCLUDE_PACKAGE_DATA,
  package_dir = PACKAGE_DIR,
  install_requires=INSTALL_REQUIRES,
  cmdclass=CMDCLASS,
  entry_points=ENTRY_POINTS
)

    

    

     





def convert_types(correct_types):
  converted_types = []
  for type in correct_types:
    if type == int or type == float:
      converted_types.append("c") # continuous if a number
    else:
      converted_types.append("d") # discrete if a string
  return converted_types

def generate_meta_data(headers, response_header, features_to_ignore):
  meta_data = []
  for feature in headers:
    if feature == response_header:
      meta_data.append("class")
    elif feature in features_to_ignore:
      meta_data.append("ignore")
    else:
      meta_data.append("")
  return meta_data

def generate_tab_data(headers, types, meta_data, data, output):
  tab_data = open(output, 'w')
  tab_data.write("\t".join(headers)+'\n')
  tab_data.write("\t".join(types)+'\n')
  tab_data.write("\t".join(meta_data)+'\n')

  for row in data:
    tab_data.write("\t".join(str(x) for x in row)+'\n')

def merge_data(orig_data, obscured_data, header, obscured_tag):
  merged_data = []

  obscured_header = [feature+obscured_tag for feature in header]
  merged_header = [None]*(len(header)+len(obscured_header))
  merged_header[::2] = header
  merged_header[1::2] = obscured_header
  merged_data.append(merged_header)
 
  for row_number, orig_row in enumerate(orig_data):
    obscured_row = obscured_data[row_number+1]
    merged_row = [None]*(len(orig_row)+len(obscured_row))
    merged_row[::2] = orig_row
    merged_row[1::2] = obscured_row
    merged_data.append(merged_row)

  return merged_data

def load(audit_params, OUTPUT_DIR):
  orig_train, orig_test, obscured_test, headers, response_header, features_to_ignore, correct_types, obscured_tag = audit_params

  # generate data needed for finding CN2 rules
  converted_types = convert_types(correct_types)
  meta_data = generate_meta_data(headers, response_header, features_to_ignore)
  
  orig_train_tab = OUTPUT_DIR+"/original_train.tab"
  generate_tab_data(headers, converted_types, meta_data, orig_train, orig_train_tab)

  orig_test_tab = OUTPUT_DIR+"/original_test.tab"
  generate_tab_data(headers, converted_types, meta_data, orig_test, orig_test_tab)
  
  # generate merged_data
  merged_data = merge_data(orig_test, obscured_test, headers, obscured_tag)

  return orig_train_tab, orig_test_tab, merged_data



import json
import os, shutil
import random
import csv

import matplotlib
matplotlib.use('Agg') # Set the back-end
import matplotlib.pyplot as plt


def load_audit_confusion_matrices(filename):
  """
  Loads a confusion matrix in a two-level dictionary format.

  For example, the confusion matrix of a 75%-accurate model
  that predicted 15 values (and mis-classified 5) may look like:
  {"A": {"A":10, "B": 5}, "B": {"B":5}}

  Note that raw boolean values are translated into strings, such that
  a value that was the boolean True will be returned as the string "True".
  """

  with open(filename) as audit_file:
    audit_file.next() # Skip the first line.

    # Extract the confusion matrices and repair levels from the audit file.
    confusion_matrices = []
    for line in audit_file:
      separator = ":"
      separator_index = line.index(separator)

      comma_index = line.index(',')
      repair_level = float(line[separator_index+2:comma_index])
      raw_confusion_matrix = line[comma_index+2:-2]
      confusion_matrix = json.loads( raw_confusion_matrix.replace("'","\"") )
      confusion_matrices.append( (repair_level, confusion_matrix) )

  # Sort the repair levels in case they are out of order for whatever reason.
  confusion_matrices.sort(key = lambda pair: pair[0])
  return confusion_matrices


def graph_audit(filename, measurers, output_image_file):
  with open(filename) as audit_file:
    header_line = audit_file.readline()[:-1] # Remove the trailing endline.

  confusion_matrices = load_audit_confusion_matrices(filename)

  x_axis = [repair_level for repair_level, _ in confusion_matrices]
  y_axes = []

  # Graph the results for each requested measurement.
  for measurer in measurers:
    y_axis = [measurer(matrix) for _, matrix in confusion_matrices]
    plt.plot(x_axis, y_axis, label=measurer.__name__)
    y_axes.append(y_axis)

  # Format and save the graph to an image file.
  plt.title(header_line)
  plt.axis([0,1,0,1.1]) # Make all the plots consistently sized.
  plt.xlabel("Repair Level")
  plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
  plt.savefig(output_image_file, bbox_inches='tight')
  plt.clf() # Clear the entire figure so future plots are empty.

  # Save the data used to generate that image file.
  with open(output_image_file + ".data", "w") as f:
    writer = csv.writer(f)
    headers = ["Repair Level"] + [calc.__name__ for calc in measurers]
    writer.writerow(headers)
    for i, repair_level in enumerate(x_axis):
      writer.writerow([repair_level] + [y_vals[i] for y_vals in y_axes])

def graph_audits(filenames, measurer, output_image_file):
  features = []
  y_axes = []
  for filename in filenames:
    with open(filename) as audit_file:
      header_line = audit_file.readline()[:-1] # Remove the trailing endline.
      feature = header_line[header_line.index(":")+1:]

    confusion_matrices = load_audit_confusion_matrices(filename)
    x_axis = [repair_level for repair_level, _ in confusion_matrices]
    y_axis = [measurer(matrix) for _, matrix in confusion_matrices]
    plt.plot(x_axis, y_axis, label=feature)

    features.append(feature)
    y_axes.append(y_axis)

  # Format and save the graph to an image file.
  plt.title(measurer.__name__)
  plt.axis([0,1,0,1.1]) # Make all the plots consistently sized.
  plt.xlabel("Repair Level")
  plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
  plt.savefig(output_image_file, bbox_inches='tight')
  plt.clf() # Clear the entire figure so future plots are empty.

  # Save the data used to generate that image file.
  with open(output_image_file + ".data", "w") as f:
    writer = csv.writer(f)
    headers = ["Repair Level"] + features
    writer.writerow(headers)
    for i, repair_level in enumerate(x_axis):
      writer.writerow([repair_level] + [y_vals[i] for y_vals in y_axes])


def rank_audit_files(filenames, measurer):
  scores = []
  for filename in filenames:
    with open(filename) as audit_file:
      header_line = audit_file.readline()[:-1] # Remove the trailing endline.
      feature = header_line[header_line.index(":")+1:]

    confusion_matrices = load_audit_confusion_matrices(filename)
    _, start_matrix = confusion_matrices[0]
    _, end_matrix = confusion_matrices[-1]
    score_difference = measurer(start_matrix)-measurer(end_matrix)
    scores.append( (feature, score_difference) )

  scores.sort(key = lambda score_tup: score_tup[1], reverse=True)
  return scores


def group_audit_ranks(filenames, measurer, similarity_bound=0.05):
  """
  Given a list of audit files, rank them using the `measurer` and
  return the features that never deviate more than `similarity_bound`
  across repairs.
  """

  def _partition_groups(feature_scores):
    groups = []
    for feature, score in feature_scores:
      added_to_group = False

      # Check to see if the feature belongs in a group with any other features.
      for i, group in enumerate(groups):
        mean_score, group_feature_scores = group
        if abs(mean_score - score) < similarity_bound:
          groups[i][1].append( (feature, score) )

          # Recalculate the representative mean.
          groups[i][0] = sum([s for _, s in group_feature_scores])/len(group_feature_scores)
          added_to_group = True
          break

      # If this feature did not much with the current groups, create another group.
      if not added_to_group:
        groups.append( [score, [(feature,score)]] )

    # Return just the features.
    return [[feature for feature, score in group] for _, group in groups]


  score_dict = {}
  features = []
  for filename in filenames:
    with open(filename) as audit_file:
      header_line = audit_file.readline()[:-1] # Remove the trailing endline.
      feature = header_line[header_line.index(":")+1:]
      features.append(feature)

    confusion_matrices = load_audit_confusion_matrices(filename)
    for rep_level, matrix in confusion_matrices:
      score = measurer(matrix)
      if rep_level not in score_dict:
        score_dict[rep_level] = {}
      score_dict[rep_level][feature] = score

  # Sort by repair level increasing repair level.
  score_keys = sorted(score_dict.keys())

  groups = [features]
  while score_keys:
    key = score_keys.pop()
    new_groups = []
    for group in groups:
      group_features = [(f, score_dict[key][f]) for f in group]
      sub_groups = _partition_groups(group_features)
      new_groups.extend(sub_groups)
    groups = new_groups

  return groups


def test():
  TMP_DIR = "tmp"
  if not os.path.exists(TMP_DIR):
    os.makedirs(TMP_DIR)

  test_contents = "GFA Audit for: Test Feature\n0.0.data:[0.0, {'A': {'B': 100}, 'B': {'B': 199}}]\n0.1.data:[0.1, {'A': {'B': 100}, 'B': {'B': 199}}]\n0.5.data:[0.5, {'A': {'B': 100}, 'B': {'B': 199}}]\n1.0.data:[1.0, {'A': {'B': 100}, 'B': {'B': 199}}]\n"
  test_filenames = [TMP_DIR + "/test_audit_1.audit",
                    TMP_DIR + "/test_audit_2.audit"]

  # Prepare the sample audit files.
  for filename in test_filenames:
    with open(filename, "w") as f:
      f.write(test_contents)

  # A mock measurement measurer that returns a random number.
  def mock_measurer(conf_matrix):
    return random.random()

  # Perform the audit and save it an output image.
  measurers = [mock_measurer, mock_measurer]
  output_image = TMP_DIR + "/test_image.png"
  graph_audit(test_filenames[0], measurers, output_image) # Only need to test 1.

  file_not_empty = os.path.getsize(output_image) > 0
  print "image file generated? --", file_not_empty

  file_not_empty = os.path.getsize(output_image + ".data") > 0
  print "data file generated? --", file_not_empty

  ranked_features = rank_audit_files(test_filenames, mock_measurer)
  print "ranked features sorted? --", ranked_features[0] > ranked_features[1]

  output_image = TMP_DIR + "/test_image2.png"
  graph_audits(test_filenames, mock_measurer, output_image)
  file_not_empty = os.path.getsize(output_image) > 0
  print "ranked image file generated? --", file_not_empty

  shutil.rmtree(TMP_DIR)

if __name__=="__main__":
  test()



import numpy

def FreedmanDiaconisBinSize(feature_values):
  """
  The bin size in FD-binning is given by size = 2 * IQR(x) * n^(-1/3)
  More Info: https://en.wikipedia.org/wiki/Freedman%E2%80%93Diaconis_rule

  If the BinSize ends up being 0 (in the case that all values are the same),
  return a BinSize of 1.
  """

  q75, q25 = numpy.percentile(feature_values, [75, 25])
  IQR = q75 - q25

  return 2.0 * IQR * len(feature_values) ** (-1.0/3.0)


def test():
  values = range(0,100)
  bin_size = FreedmanDiaconisBinSize(values)
  correct_bin_size = 21
  bin_size = round(bin_size)
  print "FreedmanDiaconisBinSize -- correct size of bins? ", bin_size == correct_bin_size


if __name__=="__main__":
  test()



from splitters import split_by_percent
import csv
import random

train_percentage = 2.0/3.0
filename = "test_data/german_categorical.csv"
max_entries = None
correct_types = [str, int, str, str, int, str,str,
                 int,str,str,int, str, int, str, str, str, int,
                 str, int, str, str, str]

AGE_COL = 12

def load_data():
  with open(filename) as f:
    reader = csv.reader(f)
    data = [row for row in reader]
    headers = data.pop(0)

    if max_entries:
      data = random.sample(data, max_entries)

    for i, row in enumerate(data):
      for j, correct_type in enumerate(correct_types):
        data[i][j] = correct_type(row[j])

    train, test = split_by_percent(data, train_percentage)

  return headers, train, test


def test():
  headers, train, test = load_data()
  print "load_data unpacks correctly? -- ", (headers != None and train != None and test != None)

  gathered_types = []
  for i, header in enumerate(headers):
    if all( isinstance(row[i],float) for row in train + test ):
      gathered_types.append(float)
    elif all( isinstance(row[i],int) for row in train + test ):
      gathered_types.append(int)
    elif all( isinstance(row[i],str) for row in train + test ):
      gathered_types.append(str)

  # The below test is supposed to fail because the age column type changes from the
  # read in correct type of int to the old/young type of str
  # print "load_data types are correct? -- ", gathered_types == correct_types
  print "all headers given types? -- ", len(headers) == len(correct_types)


if __name__=="__main__":
  test()





# NOTE:These settings and imports should be the only things that change
#       across experiments on different datasets
# TODO: Make this file generalizable to all datasets

from experiments.arrests.load_data import load_data
from repairers.CategoricRepairer import Repairer
import numpy as np
import os

import matplotlib
matplotlib.use('Agg') # Set the back-end
import matplotlib.pyplot as plt

FIGURES_DIR = "figures"
if not os.path.exists(FIGURES_DIR):
  os.makedirs(FIGURES_DIR)

def run():
  feature_to_repair = 0
  repair_level = 1.0
  headers, train_data, test_data = load_data()
  orig_data = train_data
  repairer = Repairer(orig_data, feature_to_repair, repair_level)
  repaired_data = repairer.repair(test_data)

  features_to_graph = range(1, 13)
  for feature_to_graph in features_to_graph:
    header = headers[feature_to_graph]

    orig_groups = {}
    group_indices = {}
    for i, row in enumerate(orig_data):
      stratified_val = row[feature_to_repair]
      feature_val = row[feature_to_graph]
      if not stratified_val in orig_groups:
        orig_groups[stratified_val] = []
        group_indices[stratified_val] =[]
      orig_groups[stratified_val].append(feature_val)
      group_indices[stratified_val].append(i)
    rep_groups = {group:[repaired_data[i][feature_to_graph] for i in indices] for group, indices in group_indices.items()}

    data_dict = {0: {}, 1: {}}
    data_list = {0: {}, 1: {}}
    for group, data in orig_groups.items():
      data_dict[0][group] = {value: 0 for value in data}
      data_dict[1][group] = {value: 0 for value in data}
      for value in data:
        data_dict[0][group][value] += 1
      rep_data = rep_groups[group]
      for value in rep_data:
        if value in data_dict[1][group]:
          data_dict[1][group][value] += 1

    for group in orig_groups:
      data_list[0][group] = []
      data_list[1][group] = []
      for value in data_dict[0][group]:
        count0 = data_dict[0][group][value]
        count1 = data_dict[1][group][value]
        data_list[0][group].append(count0)
        data_list[1][group].append(count1)
      categories =  [value for value in data_dict[0][group]]
      n_categories = len(categories)

      count_group_orig = data_list[0][group]
      count_group_repaired = data_list[1][group]

      fig, ax = plt.subplots()

      index = np.arange(n_categories)
      bar_width = 0.35

      opacity = 0.4
      error_config = {'ecolor': '0.3'}

      rects1 = plt.bar(index, count_group_orig, bar_width,
                       alpha=opacity,
                       color='b',
                       label='Original')

      new_index = [i + bar_width for i in index]
      rects2 = plt.bar(new_index, count_group_repaired, bar_width,
                       alpha=opacity,
                       color='r',
                       label='Repaired')

      plt.xlabel('Categories')
      plt.ylabel('Count')
      plt.title(group + ' distribution over categories for feature ' + header)
      plt.xticks(index + bar_width, categories, rotation='vertical')
      plt.legend()

      plt.tight_layout()
      plt.savefig("figures/"+ header+ "_" + group + ".png")
      plt.savefig("{}/{}_{}.png".format(FIGURES_DIR, header, group))
      plt.clf()


if __name__=="__main__":
  run()




"""
To reproduce results from the "Auditing Black-box Models for Indirect Influence" 
paper (11/30/2016), set model_options in main.py to {'reproduce':True} 
"""







from AbstractModelFactory import AbstractModelFactory
from AbstractModelVisitor import AbstractModelVisitor

import os
import numpy as np
import random as rand
import tensorflow as tf
import copy
import time

TMP_DIR_HOST = 'tmp'

date_stamp = time.strftime('%a.%d.%b_%S.%M.%H', time.localtime())
log_dir = '{}/{}__{}'.format(TMP_DIR_HOST, date_stamp, rand.randint(1,1000000))
os.makedirs(log_dir)  


class ModelFactory(AbstractModelFactory):
  def __init__(self, *args, **kwargs):
    self.num_epochs = 500          # number of passes through training data
    self.batch_size = 20            # number of points trained on per step
    self.learning_rate = .01            # learning rate for optimizer
    self.beta1_decay = 0.95              # see Tensorflow's documentation for tf.train.AdamOptimizer
    self.beta2_decay = 0.99                # for information on beta decay
    self.epsilon = 1e-8                 # small number for stability of optimizer
    self.init_shuffle = True           # shuffle data before training starts
    self.iter_shuffle = True           # shuffle batch for each step
    self.feats_to_ignore = []           # features to ignore during training
    self.hidden_layer_sizes = []        # if empty, no hidden layers are used
    self.constant_descent = False       # flag to implement tf.GradientDescentOptimizer

    # Set manual settings.
    if 'options' in kwargs:
      options = kwargs['options']
      if 'num_epochs' in options:
        self.num_epochs = options.pop('num_epochs')
      if 'batch_size' in options:
        self.batch_size = options.pop('batch_size')
      if 'learning_rate' in options:
        self.learning_rate = options.pop('learning_rate')
      if 'beta1_decay' in options:
        self.beta1_decay = options.pop('beta1_decay')
      if 'beta2_decay' in options:
        self.beta2_decay = options.pop('beta2_decay')
      if 'epsilon' in options:
        self.epsilon = options.pop('epsilon')
      if 'init_shuffle' in options:
        self.init_shuffle = options.pop('init_shuffle')
      if 'iter_shuffle' in options:
        self.iter_shuffle = options.pop('iter_shuffle')
      if 'feats_to_ignore' in options:
        self.feats_to_ignore = options.pop('feats_to_ignore')
      if 'hidden_layer_sizes' in options:
        self.hidden_layer_sizes = options.pop('hidden_layer_sizes')
      if 'reproduce' in options:
        reproduce = options.pop('reproduce')
        if reproduce:
          self.num_epochs = 100
          self.batch_size = 100
          self.learning_rate = 0.01
          self.hidden_layer_sizes = []
          self.init_shuffle = False
          self.iter_shuffle = False
          self.constant_descent = True

    # Initiate inheritance
    super(ModelFactory, self).__init__(*args, **kwargs)
    self.verbose_factory_name = 'TensorFlow_Network'


    # Maps each header to all possible values for it's column.
    self.col_vals = {header:{row[i] for row in self.all_data} for i, header in enumerate(self.headers)}
    self.response_index = self.headers.index(self.response_header)
    self.num_outcomes = len(self.col_vals[self.response_header])

    # Mark any categorical features for column expansion.
    # Categorical features are transfered into -1, 1 binary features for each value.
    self.columns_to_expand = []
    for i, header in enumerate(self.headers):
      categorical = any(type(val)==str for val in self.col_vals[header])
      if i == self.response_index or categorical:
        self.columns_to_expand.append(header)
    
    # All non-categorical features will be standardized (mean=0, variance=1).
    self.standardizers = {}
    for i, header in enumerate(self.headers):
      if any (type(row[i])==str for row in self.all_data):
        mean = None
        std_dev = None
      else:
        mean = sum(row[i] for row in self.all_data)/float(len(self.all_data))
        std_dev = np.sqrt(sum(((row[i] - mean) ** 2) for row in self.all_data)/float(len(self.all_data)))
      self.standardizers[header] = {'mean':mean, 'std_dev':std_dev}
    
    # If the response column is shifted by expanding categorical features, 
    # update the response index.
    response_col_shift = 0
    for header in self.headers[:self.response_index]:
      if header in self.columns_to_expand:
        response_col_shift += len(self.col_vals[header]) - 1
    self.adjusted_response_index = self.response_index + response_col_shift
    
    # Map each outcome to an index
    self.outcome_trans_dict = {val:i for i, val in enumerate(self.col_vals[self.response_header])}

  def build(self, train_set):
    model_name="{}/{}_{}_{}.model".format(log_dir, self.verbose_factory_name, self.factory_name, time.time())
    # prepare train data for learning
    if self.init_shuffle == True:
      np.random.seed(123)
      np.random.shuffle(train_set)
    expanded_and_stdized_train_set, self.expanded_headers = expand_and_standardize_dataset(self.response_index, self.response_header, train_set, self.col_vals, self.headers, self.standardizers, self.feats_to_ignore, self.columns_to_expand, self.outcome_trans_dict)
    train_matrix, train_outcomes = list_to_tf_input(expanded_and_stdized_train_set, self.adjusted_response_index, self.num_outcomes)
    train_size, num_features = train_matrix.shape
    
    # Ensure there are no pre-existing nodes in the tensorflow graph.
    tf.reset_default_graph()
    
    # input and output nodes
    with tf.name_scope('input'):
      x = tf.placeholder(tf.float32, shape=[None, num_features], name='x_input') # Input
      y_ = tf.placeholder(tf.float32, shape=[None, self.num_outcomes], name='y_input') # Correct Output
    
    # Construct the layer architecture.
    # Generate a layer for the input and for each additional hidden layer.
    layer_sizes = [num_features] + self.hidden_layer_sizes + [self.num_outcomes]
    layers = [x] # Count the input as the first layer
    for i in xrange(len(layer_sizes) - 1):
      layer_size = layer_sizes[i]
      prev_layer = layers[-1]
      next_layer_size = layer_sizes[i+1]
      
      # Create a new layer with initially random weights and biases.  
      if i == len(layer_sizes) - 1:   # Final layer is output.
        layer_name = 'output'
      else:
        layer_name = 'hidden_layer_{}'.format(i)
      with tf.name_scope(layer_name):
        tf.set_random_seed(123)
        W = tf.Variable(tf.random_normal([layer_size, next_layer_size]), name='weights')
        b = tf.Variable(tf.random_normal([next_layer_size]), name='biases')
        new_layer = (tf.add(tf.matmul(prev_layer, W),b))  # Wx + b
        layers.append(new_layer)

    y = layers[-1]

    # Optimization.
    with tf.name_scope('accuracy'):
      correct_prediction = tf.equal(tf.argmax(y,1), tf.argmax(y_,1))
      accuracy = tf.reduce_mean(tf.cast(correct_prediction, tf.float32))  
    with tf.name_scope('cross_entropy'):
      cross_entropy = tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits(logits=y, labels=y_))
    with tf.name_scope('train_step'):
      if self.constant_descent:
        train_step = tf.train.GradientDescentOptimizer(learning_rate = self.learning_rate, name='Gradient_Descent').minimize(cross_entropy)
      else:
        train_step = tf.train.AdamOptimizer(learning_rate=self.learning_rate, beta1=self.beta1_decay, beta2=self.beta2_decay, epsilon=self.epsilon, name='Adam_Optimizer').minimize(cross_entropy)
      

    # Summaries for TensorBoard
    tf.summary.scalar('accuracy', accuracy)
    tf.summary.scalar('entropy', cross_entropy)

    saver = tf.train.Saver()
    merge = tf.summary.merge_all()
    
    with tf.Session() as sess:

      writer = tf.summary.FileWriter(log_dir, graph=sess.graph)
      tf.global_variables_initializer().run()
      num_steps = self.num_epochs * train_size // self.batch_size
      for step in xrange(num_steps):
        # get batch
        offset = (step * self.batch_size) % train_size
        batch_data = train_matrix[offset:(offset + self.batch_size), :]
        batch_labels = train_outcomes[offset:(offset + self.batch_size)]

        # shuffle
        if self.iter_shuffle == True:
          np.random.seed(step)
          rng_state = np.random.get_state()
          np.random.shuffle(batch_data)
          np.random.set_state(rng_state)
          np.random.shuffle(batch_labels)
        # train and write summary
        _, summary = sess.run([train_step, merge], feed_dict={x:batch_data, y_:batch_labels})
        writer.add_summary(summary, global_step=step+1)
        # Save checkpoints every 1000 steps
        if step % 50 == 0:
          model_name="{}/{}_{}_{}.model".format(log_dir, self.verbose_factory_name, self.factory_name, time.time())
          checkpoint = saver.save(sess, model_name, global_step=step+1)

      # Save checkpoint upon completion of training
      model_name="{}/{}_{}_{}.model".format(log_dir, self.verbose_factory_name, self.factory_name, time.time())
      checkpoint = saver.save(sess, model_name, global_step=num_steps)

      # print 'Train Accuracy:', accuracy.eval(feed_dict={x:train_matrix, y_:train_outcomes})

    return ModelVisitor(model_name, checkpoint, saver, self.response_header, self.response_index, self.adjusted_response_index, self.num_outcomes, x, y_, y, self.outcome_trans_dict, self.headers, self.expanded_headers, self.standardizers, self.col_vals, self.feats_to_ignore, self.columns_to_expand)


class ModelVisitor(AbstractModelVisitor):

  def __init__(self, model_name, checkpoint, model_saver, response_header, response_index, adjusted_response_index, num_outcomes, x, y_, y, outcome_trans_dict, headers, expanded_headers, standardizers, train_col_vals, feats_to_ignore, columns_to_expand):
    super(ModelVisitor,self).__init__(model_name)
    self.model_saver = model_saver
    self.checkpoint = checkpoint
    self.response_index = response_index
    self.response_header = response_header
    self.adjusted_response_index = adjusted_response_index
    self.num_outcomes = num_outcomes
    self.x = x
    self.y_ = y_
    self.y = y
    self.outcome_trans_dict = outcome_trans_dict
    self.headers = headers
    self.expanded_headers = expanded_headers
    self.train_expanded_headers = expanded_headers
    self.train_col_vals = train_col_vals
    self.standardizers = standardizers
    self.feats_to_ignore = feats_to_ignore
    self.columns_to_expand = columns_to_expand

  def test(self, test_set, test_name=""):
    expanded_and_stdized_test_set, self.test_set_expanded_headers = expand_and_standardize_dataset(self.response_index, self.response_header, test_set, self.train_col_vals, self.headers, self.standardizers, self.feats_to_ignore, self.columns_to_expand, self.outcome_trans_dict)
    if self.test_set_expanded_headers != self.train_expanded_headers:
      raise ValueError('Feature dimensions do not align!')
    else:
      test_matrix, test_labels = list_to_tf_input(expanded_and_stdized_test_set, self.adjusted_response_index, self.num_outcomes)

    with tf.Session() as sess:
   
      self.model_saver.restore(sess, self.checkpoint)
      predictions = tf.argmax(self.y, 1).eval(feed_dict={self.x: test_matrix, self.y_:test_labels}, session=sess)

    predictions_dict = {i:key for key,i in self.outcome_trans_dict.items()}
    predictions = [predictions_dict[pred] for pred in predictions]

    return zip([row[self.response_index] for row in test_set], predictions)

def list_to_tf_input(data, response_index, num_outcomes):
  """
  Separates the outcome feature from the data and creates the onehot vector for each row.
  """
  matrix = np.matrix([row[:response_index] + row[response_index+1:] for row in data])
  outcomes = np.asarray([row[response_index] for row in data], dtype=np.uint8)
  outcomes_onehot = (np.arange(num_outcomes) == outcomes[:, None]).astype(np.float32)

  return matrix, outcomes_onehot

def expand_and_standardize_dataset(response_index, response_header, data_set, col_vals, headers, standardizers, feats_to_ignore, columns_to_expand, outcome_trans_dict):
  """
  Standardizes continuous features and expands categorical features.
  """
  # expand and standardize
  modified_set = []
  for row_index, row in enumerate(data_set):
    new_row = []
    for col_index, val in enumerate(row):
      header = headers[col_index]

      # Outcome feature -> index outcome
      if col_index == response_index:
        new_outcome = outcome_trans_dict[val]
        new_row.append(new_outcome)

      # Ignored feature -> pass
      elif header in feats_to_ignore:
        pass
      
      # Categorical feature -> create new binary column for each possible value of the column
      elif header in columns_to_expand:
        for poss_val in col_vals[header]:
          if val == poss_val:
            new_cat_val = 1.0
          else:
            new_cat_val = -1.0
          new_row.append(new_cat_val)

      # Continuous feature -> standardize value with respect to its column
      else:
        new_cont_val = float((val - standardizers[header]['mean']) / standardizers[header]['std_dev'])
        new_row.append(new_cont_val)

    modified_set.append(new_row)

  # update headers to reflect column expansion
  expanded_headers = []
  for header in headers:
    if header in feats_to_ignore:
      pass
    elif (header in columns_to_expand) and (header is not response_header):
      for poss_val in col_vals[header]:
        new_header = '{}_{}'.format(header,poss_val)
        expanded_headers.append(new_header)
    else:
      expanded_headers.append(header)

  return modified_set, expanded_headers

def test():
  test_expand_and_standardize_dataset()
  test_unseen_categorical_feature()
  test_categorical_model()
  test_categorical_response()
  test_list_to_tf_input()
  test_basic_model()


def test_expand_and_standardize_dataset():
  """
  Tests if the function correctly converts categorical
  features into onehot features, correctly standardizes 
  continuous features, and correctly converts the response
  feature to indexes.
  """
  headers = ['cont_feature', 'cat_feature', 'ignored', 'response']
  data = [[75., 'A', 'ingore1', 'X'], [25, 'B', 'ignore2', 'Y'], [75, 'C', 'ignore3', 'Z'], [25., 'A', 'ignore1', 'Y']]
  resp_index = 3
  resp_header = 'response'
  feats_to_ignore = 'ignored'
  col_vals = {'cont_feature':[75., 25.], 'cat_feature':['A','B','C'], 'ignored':['ignore1', 'ignore2', 'ignore3'], 'response':['X', 'Y', 'Z']}
  columns_to_expand = ['cat_feature', 'ignored', 'response']
  outcome_trans_dict = {'X':0, 'Y':1, 'Z':2}
  standardizers = {'cont_feature':{'mean':50, 'std_dev':25}}
  new_data, new_headers = expand_and_standardize_dataset(resp_index, resp_header, data, col_vals, headers, standardizers, feats_to_ignore, columns_to_expand, outcome_trans_dict)
  correct_data = [[1.0, 1.0, -1.0, -1.0, 0], [-1.0, -1.0, 1.0, -1.0, 1], [1.0, -1.0, -1.0, 1.0, 2], [-1.0, 1.0, -1.0, -1.0, 1]]
  correct_headers = ['cont_feature', 'cat_feature_A', 'cat_feature_B', 'cat_feature_C', 'response']
  print 'expanding and standardizing data correctly? --', new_data==correct_data
  print 'expanding headers correctly? --', new_headers==correct_headers

def test_unseen_categorical_feature():
  """
  Tests if the model can correctly handle unfamiliar values
  for categorical features. The model should not create new onehot 
  features for values in the test set and should instead treat
  these values as 'not any known value for that column'. 
  """
  headers = ["predictor 1", "predictor 2", "predictor3", "response"]
  response = "response"
  random_features = ['H', 5.5, 'I']
  train_set = [[random_features[i%3],'B',1,'C'] for i in range(1,50)] + [['A','A',-1,'D'] for i in range(1,50)]
  train_set_copy = copy.copy(train_set)
  new_features = ['J',9]
  test_set = [[new_features[i%2],'B',1,'C'] for i in range(1,49)] + [['A','A',-1,'D'] for i in range(1,50)]
  all_data = train_set + test_set

  factory = ModelFactory(all_data, headers, response, name_prefix="test")
  model = factory.build(train_set)
  print "factory builds ModelVisitor? -- ", isinstance(model, ModelVisitor)

  predictions = model.test(test_set)
  print "handling unseen categorical values correctly? -- ", all([pred[0] == pred[1] for pred in predictions])


def test_list_to_tf_input():
  data = [[0,0],[0,1],[0,2]]
  tf_matrix, tf_onehot = list_to_tf_input(data, 1, 3)
  correct_matrix = [[0],[0],[0]]
  correct_onehot = [[1,0,0], [0,1,0], [0,0,1]]
  print "list_to_tf_input matrix correct? --",np.array_equal(tf_matrix, correct_matrix)
  print "list_to_tf_input onehot correct? --",np.array_equal(tf_onehot, correct_onehot)

def test_basic_model():
  headers = ["predictor 1", "predictor 2", "response"]
  response = "response"
  train_set = [[i,0,0] for i in range(1,50)] + [[0,i,1] for i in range(1,50)]
  test_set = [[i,0,0] for i in range(1,50)] + [[0,i,1] for i in range(1,50)]
  all_data = train_set + test_set

  factory = ModelFactory(all_data, headers, response, name_prefix="test")

  model = factory.build(train_set)

  predictions = model.test(test_set)
  intended_predictions = [pred[0] for pred in predictions]
  actual_predictions = [pred[1] for pred in predictions]
  print "predicting numeric categories correctly? -- ", actual_predictions == intended_predictions

def test_categorical_response():
  headers = ["predictor 1", "predictor 2", "response"]
  response = "response"
  train_set = [[i,0,"A"] for i in range(1,50)] + [[0,i,"B"] for i in range(1,50)]
  test_set = [[i,0,"A"] for i in range(1,50)] + [[0,i,"C"] for i in range(1,50)]
  train_set_copy = copy.copy(train_set)
  test_set_copy = copy.copy(test_set)
  all_data = train_set + test_set

  factory = ModelFactory(all_data, headers, response, name_prefix="test")
  model = factory.build(train_set)

  predictions = model.test(test_set)
  resp_index = headers.index(response)
  intended_predictions = [(test_row[resp_index], train_row[resp_index]) for train_row, test_row in zip(train_set_copy,test_set_copy)]
  print "predicting string-categories correctly? -- ", predictions == intended_predictions

def test_categorical_model():
  """
  Tests whether the model can correctly learn on the train set. 
  The model should have poor accuracy on the test set
  because the outcomes do not match those of the train set.
  The model should incorrectly predict response 'B' for predictor 'B'
  when response 'C' is the correct response.
  """
  headers = ["predictor", "response"]
  response = "response"
  train_set = [["A","A"] for i in range(1,50)] + [["B","B"] for i in range(1,50)]
  train_set_copy = copy.copy(train_set)
  test_set = [["A","A"] for i in range(1,50)] + [["B","C"] for i in range(1,50)]
  all_data = train_set + test_set
  factory = ModelFactory(all_data, headers, response, name_prefix="test", options={})

  model = factory.build(train_set)

  predictions = model.test(test_set)
  resp_index = headers.index(response)
  intended_predictions = [(test_row[resp_index], train_row[resp_index]) for train_row, test_row in zip(train_set_copy,test_set)]
  print "predicting string-categories correctly? -- ", predictions == intended_predictions

if __name__=='__main__':
  test()
